//picpid.c
/*
	This is a simple but complete PID velocity controller test scaffold.  

	This version does max speed calibration at the beginning of the cycle.
	Any "I" value  works great for constant speed holding.
	
	This version handles ramping for constant speed quite well.  Multiple ramping rates can be
	used as well and it handles direction changes too.  Changes from fast to very slow in the
	same direction cause hesitation as the error rates slam together though.  
	
	This version is running at 1Khz and doubles the values read by the ADC.  Response is MUCH
	better and you can use the "I" value in the holding mode.  Doubling the ADC values takes into
	consideration that I'm not seeing more than 30% of the max voltage as CEMF.  I'm amplifying 
	in software instead of hardware.
	
	There are still a few bugs in here where I do not do boundary checking.  See if you can find them!
	9/15/2008
	
	PID
	t = current time
	error(t) = Vdesired(t) - Vmeasured(t)
	Vcontrol(t) = Vprop(t) + Vinteg(t) + Vderiv(t)
	Vprop(t) = Kp * error(t) (Kp = proportional gain)
	Vinteg(t) = Ki * SUM(error(0 to t)) (Ki = integral gain; SUM = sum of all past errors)
	Vderiv(t) = Kd * (error(t) - error(t-1)) (Kd = deriv gain; error difference between last and now)
*/
#include <16F73.h>
#fuses HS,NOWDT,NOPROTECT,NOBROWNOUT
#use delay(clock=14740000)
#use rs232(baud=115200, xmit=PIN_C6, rcv=PIN_C7)	//hardware UART
#use FAST_IO(C)
#use FAST_IO(B)
#use FAST_IO(A)

//If you make this 32 you get more range, but bigger and slower code.
#type long=16

#define PWM PIN_C2
#define DIR1A PIN_B0
#define DIR1B PIN_B1

//*******************************************************************************************
// GLOBALS
//*******************************************************************************************
signed long ehist;										//error history accumulated
signed long calcPos;									//desired velocity (arbitrary units)
signed long Verror;
signed long lerror;										//last error value
signed long currPos;									//where we are now
signed long Vc, Vp, Vi, Vd;
int action;												//what to do
signed int Kp = 20;										//proportional gain
signed int Ki = 2;										//integral gain
signed int Kd = 5;										//derivitive gain

signed long mV;											//max velocity
signed int acc;											//Acceleration (ramp tics)
signed int vAcc;										//ramp acceleration increments
signed long endPos;										//actual final velocity or location

enum {
	A_HOLD,												//Hold current position
	A_CONST,											//Hold current velocity
	A_RAMP,												//Move to next trapazoidal speed increment
	A_STOP
};

signed long pwmval;

//*******************************************************************************************
// FUNCTIONS
//*******************************************************************************************
void init(void)
{
	/*
	A0 = AN0 analog input for direction A side of motor
	A1 = AN1 analog input for direction B side of motor
	A2 = digital output (unused)
	A3 = AN3 analog input for speed pot input
	A4 = digital output (unused)
	A5 = digital output (unused)
	*/
	set_tris_a(0b00001011);
	setup_adc(ADC_CLOCK_DIV_32);
	setup_adc_ports(RA0_RA1_RA3_ANALOG);
	/*
	B0 = DIR1A output
	B1 = DIR2A output
	B2-B5 unused outputs
	B6 = PGD for ICSP
	B7 = PGC for ICSP
	*/
	set_tris_b(0x00);
	
	/*
	C0 = unused output
	C1 = PWM2 output (unused)
	C2 = PWM1 output
	C3 = unused output
	C4 = unused output
	C5 = unused output
	C6 = TxD output
	C7 = RxD input
	*/
	set_tris_c(0b10000000);
	
	setup_timer_2(T2_DIV_BY_16,255,1);					//about 900 Hz PWM
	setup_ccp1(CCP_PWM_PLUS_3);							//full 10 bits, sets 2 LSB = 1 too
	
	ehist = 0;											//no history yet
	lerror = 0;
}

signed long getError()
/*
 * Find the difference between where we want to be and where we are.
 */
{
	signed long error;
	unsigned int16 ma,mb;
	
	setup_ccp1(CCP_OFF);							//Turn off PWM	
	delay_us(500);									//wait for steady state
	
	set_adc_channel(0);								//get forward side voltage
	delay_us(20);
	ma = read_adc();
	ma <<= 1;										//This "amplifies" the signal, making it more relevent

	set_adc_channel(1);								//get reverse side voltage
	delay_us(20);
	mb = read_adc();
	mb <<= 1;										//This "amplifies" the signal, making it more relevent
			
	currPos = (signed long)((signed long)ma - (signed long)mb);

	//printf("ma= %lu  mb= %lu\n\r",ma,mb);

	setup_ccp1(CCP_PWM_PLUS_3);
	error = calcPos - currPos;

	return error;
}

signed long getCorrection(signed long Vdesired)
/*
 * Returns the Vcontrol(t) value from all gains and readings.  This is what calculates
 * The P, I and D values and combines them for a single correctional value.
 */
{
	signed long Vcontrol,Vderiv;
	signed int32 temp;									//way big so it can't be exceeded.
	
	Verror = getError();								//Get error
	Vderiv = Verror - lerror;							//NOTE: this is typically a negative value!

	Vp =(signed long) (Kp * Verror);					//calculate Vproportional
	Vi = (signed long)(Ki * ehist);						//calculate Vintegral
	ehist += Verror;									//add to error history
	Vd = (signed long)(Kd * Vderiv);					//get incremental error from last time
	lerror = Verror;									//calculate Vderivative
	
	temp = Vp + Vi + Vd;
	if (temp > 1023)
	{
		Vcontrol = 1023;								//Need to avoid a rollover!
//		printf("over!\n\r");
	}
	else if (temp < -1023)
	{
		Vcontrol = -1023;
//		printf ("under!\n\r");
	}
	else
	{		
		Vcontrol = Vp + Vi + Vd;						//figure new control setting
	}
	
	//printf("err= %ld Vp:%ld Vi:%ld Vd:%ld\n\r",Verror,Vp,Vi,Vd);
	//printf("currPos= %ld\n\r",currPos);
	return Vcontrol;
}

void setMotor(signed long control)
/*
 * Apply the calculated changes to set the proper motor speed and direction.
 */
{	
	pwmval = abs(control);							//absolute PWM value, no signs used
	//printf("%ld\n\r",control);
	if (control > 0)
	{
		output_low(DIR1A);							//Setting my direction bits
		output_high(DIR1B);
	}
	else
	{
		output_low(DIR1B);
		output_high(DIR1A);
	}

	set_pwm1_duty(pwmval);							//set PWM value
}

void main(void)
{
	signed int nP = 0;
	signed int nI = 0;
	signed int nD = 0;
	signed long nV = 0;
	signed long t1;			//tempvar
	signed int dir = 1;		//direction setting, 1 for one direction, -1 for the other.
	char key;
	signed long maxSpeed,currMax;
	int counter;
	
	init();
		
	while (kbhit())			//Flush out the keyboard buffer so we don't get any surprises!
	{
		getch();
	}

	printf("%c[2J%c[H",27,27);						//I use Teraterm, this will clear the screen.
	printf("Calibrating for max speed value, hit space bar to start.");
	while (!kbhit());

	//Calibrate max speed
	setMotor(-1023);
	action = A_CONST;	
	calcPos = 0;
	maxSpeed = 0;
	delay_ms(100);
	for (counter=1;counter<10;counter++)			//Find the max motor speed rating that can be attained.
	{
		maxSpeed += getError();
		currMax = maxSpeed/counter;					//average readings.
		printf("%c[2J%c[H",27,27);					//home cursor in TeraTerm
		printf("max speed= %ld",currMax);
		delay_ms(500);
	}
	maxSpeed = currMax;
	mV = currMax;
	acc = 100;										//ramping rate
	vAcc = mV/acc;
	counter = vAcc +1;								//get "nearly there" value

	pwmval = 0;
	set_pwm1_duty(0);
	
	endPos = 10L;
	currPos = 0L;	
	calcPos = 0L;
	
	Kp = 5;											//proportional gain
	Ki = 0;											//integral gain
	Kd = 2;											//derivitive gain
	t1 = 0;
	
	nP = Kp;
	nI = Ki;
	nD = Kd;
	nV = endPos;
	
	while (kbhit())									//clear kbd buffer
	{
		getch();
	}

//	printf("%c[2J%c[H",27,27);						//Clear Screen/Home Cursor for Teraterm
	printf("starting up\n\r");
	printf("Velocity = %ld Kp= %d Ki= %d Kd= %d",endPos,nP,nI,nD);
	while (!kbhit());
	
	currPos = 0;
	
	while (action != A_STOP)
	{
		delay_ms(5);
						
		if (action == A_CONST)						//Figure the speed ramping value.
		{
			if (abs(endPos - calcPos) > counter)	//don't bother to increment if we are close
				calcPos += (dir*vAcc);
			else
				calcPos = endPos;
		}
		
		Vc = getCorrection(calcPos);		
		setMotor(Vc);
		
		//printf("currPos: %ld  endPos: %ld calcPos= %ld\n\r",currPos,endPos,calcPos);
		
		if (kbhit())
		{
			key = getch();
			switch (key)
			{
				case 'R':					//Reset to initial values
					Vc=0;
					Vp=0;
					Vi=0;
					Vd=0;
					currPos = 0;
					calcPos = 0;
					endPos = 10L;
					nP=5;
					nI=0;
					nD=2;
					nV=endPos;
					eHist=0;
					vError=0;
					dir = 1;
					vAcc = 1;
					counter = vAcc+1;
					break;
				case 'P':					// Increase P value
					nP++;
					break;
				case 'p':					// Decrease P value
					nP--;
					break;
				case 'I':					// Increase I value
					nI++;
					break;
				case 'i':					// Decrease I value
					nI--;
					break;
				case 'D':					// Increase D value
					nD++;
					break;
				case 'd':					//Decrease D value
					nD--;
					break;
				case 'L':					//Load the current settings
					Kp = nP;
					Ki = nI;
					Kd = nD;
					calcPos = endPos;
					endPos = nV;
					if ((endPos - calcPos) < 0)
						dir = -1;
					else
						dir = 1;
					eHist = 0;				//clear out the error so we don't have to wait.
					vError = 0;
					break;
				case 'V':					//Increase velocity by 1
					nV++;
					break;
				case 'v':					//Decrease velocity by 1
					nV--;
					break;
				case 'S':					//Stop the test and turn off
					action = A_STOP;
					break;
				case 'M':					//Increase velocity by 100
					nV+=100;
					break;
				case 'm':					//Decrease velocity by 100
					nV-=100;
					break;
				case 'a':					//Decrease the ramp rate by 1
					vAcc -= 1;
					if (vAcc < 1)
						vAcc = 1;
					counter = vAcc+1;
					break;
				case 'A':					//Increase the ramp rate by 1
					vAcc += 1;
					counter = vAcc+1;
					break;
				default:
					break;
			}
			if (nV > maxSpeed)
				nV = maxSpeed;
			else if (nV < (0-maxSpeed))
				nv = (0-maxSpeed);

			printf("%c[2J%c[H",27,27);
			if (action == A_CONST)
				printf("(%ld)Velocity= %ld  Kp= %d  Ki= %d  Kd= %d vAcc= %d\n\r",currPos,nV,nP,nI,nD,vAcc);
			else
				printf(" STOPPED");
					
		}
	}		
	setup_ccp1(CCP_OFF);
	output_low(DIR1A);
	output_low(DIR1B);
	while(1);
}
